from datetime import datetime
from typing import List, Optional
import requests
from ani_scrapy.sync_api.browser import SyncBrowser
from ani_scrapy.sync_api.scrapers.jkanime.file_link import (
    get_mediafire_file_link,
    get_streamwish_file_link,
)
import cloudscraper
from urllib.parse import quote
from bs4 import BeautifulSoup, Tag
from ani_scrapy.core.exceptions import (
    ScraperBlockedError,
    ScraperParseError,
    ScraperTimeoutError,
)
from ani_scrapy.core.schemas import (
    _AnimeType,
    _RelatedType,
    AnimeInfo,
    DownloadLinkInfo,
    EpisodeDownloadInfo,
    EpisodeInfo,
    PagedSearchAnimeInfo,
    RelatedInfo,
    SearchAnimeInfo,
)
from ani_scrapy.sync_api.base import SyncBaseScraper
from ani_scrapy.core.constants.jkanime import (
    BASE_URL,
    SEARCH_URL,
    anime_type_map,
    related_type_map,
    month_map,
)


get_file_download_link = {
    "Streamwish": get_streamwish_file_link,
    "Mediafire": get_mediafire_file_link,
}


class JKAnimeScraper(SyncBaseScraper):
    """
    Sync scraper for jkanime.net.
    """

    def __init__(self, verbose: bool = False, level: str = "INFO"):
        super().__init__(verbose, level)
        self._log("Initializing JKAnime scraper", "DEBUG")

    def _parse_anime_info(self, element: Tag) -> SearchAnimeInfo:
        try:
            anime_id = element.select_one("a[href]")["href"].split("/")[-2]
            type_ = element.select_one("li.anime").text
            title = element.select_one("h5 > a").text
            poster = element.select_one("a > div")["data-setbg"]

            self._log(f"Found anime '{title}' with id '{anime_id}'", "DEBUG")
            return SearchAnimeInfo(
                id=anime_id,
                title=title,
                poster=poster,
                type=anime_type_map.get(type_, _AnimeType.TV),
            )
        except Exception as e:
            raise ScraperParseError(e)

    def search_anime(self, query: str) -> PagedSearchAnimeInfo:
        """
        Search for anime by query.

        Parameters
        ----------
        query : str
            The query to search for.

        Returns
        -------
        PagedSearchAnimeInfo
            An object containing the page number, total pages, and animes found.

        Raises
        ------
        ValueError
            If the query is not provided or is less than 3 characters.
        ScraperBlockedError
            If the request is blocked by the server.
        ScraperTimeoutError
            If the request times out.
        ScraperParseError
            If the response from the server cannot be parsed.
        """

        if not query or len(query) < 3:
            raise ValueError(
                "The variable 'query' must be at least 3 characters"
            )

        self._log(f"Searching for anime with query '{query}'")

        safe_query = quote(query)
        search_anime_url = f"{SEARCH_URL}/{safe_query}"
        self._log(f"Using search url '{search_anime_url}'", "DEBUG")

        try:
            scraper = cloudscraper.create_scraper()
            response = scraper.get(search_anime_url)
        except requests.Timeout as e:
            raise ScraperTimeoutError("Request timed out") from e
        except Exception as e:
            raise ScraperParseError(
                f"Error fetching search results: {e}"
            ) from e

        if response.status_code == 403:
            raise ScraperBlockedError(
                f"Request failed with status code {response.status_code}"
            )
        if response.status_code == 500:
            raise ScraperTimeoutError(
                f"Request failed with status code {response.status_code}"
            )

        html_text = response.text
        soup = BeautifulSoup(html_text, "lxml")
        elements = soup.select("div.row.page_directorio > div")
        self._log(f"Found {len(elements)} animes")

        animes_info = [self._parse_anime_info(element) for element in elements]

        return PagedSearchAnimeInfo(
            page=1,
            total_pages=1,
            animes=animes_info,
        )

    def get_anime_info(
        self,
        anime_id: str,
        include_episodes: bool = True,
        tab_timeout: int = 200,
        browser: Optional[SyncBrowser] = None,
    ) -> AnimeInfo:
        """
        Get information about an anime.

        Parameters
        ----------
        anime_id : str
            The id of the anime.
        tab_timeout : int, optional
            The timeout for waiting for the tab to load. Defaults to 200.
        browser : SyncBrowser, optional
            The browser to use for scraping. If not provided, a new browser
            will be created.

        Returns
        -------
        AnimeInfo
            Information about the anime.
        """

        url = f"{BASE_URL}/{anime_id}"
        self._log(f"Getting anime info for anime with id '{anime_id}'")

        external_browser = browser is not None
        if not external_browser:
            browser = SyncBrowser().__enter__()

        page = browser.new_page()
        page.goto(url)

        html_text = page.content()
        soup = BeautifulSoup(html_text, "lxml")

        side_anime_info = soup.select_one("div.col-lg-2.picd")
        poster = side_anime_info.find("img")["src"]
        info_container = side_anime_info.select_one("div.card-bod")
        list_info = info_container.find_all("li")
        type_ = anime_type_map.get(list_info[0].text, _AnimeType.TV)
        genres = list_info[1].find_all("a")
        is_finished = None
        for l_info in list_info:
            div = l_info.find("div")
            if div:
                if div.text == "Concluido":
                    is_finished = True
                    break
                if div.text == "En emision":
                    is_finished = False
                    break

        main_anime_info = soup.select_one("div.anime_info")
        title = main_anime_info.find("h3").text
        synopsis = main_anime_info.select_one("p.scroll").text

        raw_next_episode_date = soup.select("div#proxep")
        parsed_date = None
        if raw_next_episode_date:
            next_episode_date = raw_next_episode_date[-1].text
            current_year = datetime.now().year
            parts = next_episode_date.strip().split(" ")
            day = parts[-3]
            month = month_map[parts[-1]]
            parsed_date = datetime.strptime(
                f"{current_year}-{month}-{day}", "%Y-%m-%d"
            ).date()

        select = page.query_selector("div.nice-select.anime__pagination")
        paged_episodes = select.query_selector_all("ul > li")

        all_episodes = []
        if include_episodes:
            for paged_episode in paged_episodes:
                select.click()

                page.wait_for_timeout(1000)

                current_tabs = browser.context.pages

                if len(current_tabs) > 1:
                    current_tabs[-1].close()
                    select.click()

                paged_episode.click()

                page.wait_for_timeout(tab_timeout)

                html_text = page.content()
                soup = BeautifulSoup(html_text, "lxml")
                episodes_container = soup.select_one("div#episodes-content")
                episodes = episodes_container.select("div.epcontent")
                for episode in episodes:
                    episode_number = episode.select_one("a")["href"].split(
                        "/"
                    )[-2]
                    image_preview = episode.select_one("a > div")["data-setbg"]
                    number = int(episode_number)
                    all_episodes.append(
                        EpisodeInfo(
                            number=number,
                            anime_id=anime_id,
                            image_preview=image_preview,
                        )
                    )

        navbar = page.query_selector("nav.anime-tabs.mb-4")
        options = navbar.query_selector_all("ul > li")
        options[1].click()
        html_text = page.content()
        soup = BeautifulSoup(html_text, "lxml")
        other_titles_container = soup.select_one(
            "div.rounded.bg-dark.mt-2.altert.alternativost"
        )
        all_titles = [
            t.next_sibling.text.strip()
            for t in other_titles_container.select("div > b")
        ]

        related_info_container = soup.select_one("div.col.col-lg-6")
        all_related_info = []
        child_elements = related_info_container.find_all(recursive=False)
        last_type = None
        for child_element in child_elements:
            tag = child_element.name.lower()
            if tag == "h5":
                type_text = child_element.get_text(strip=True)
                last_type = related_type_map.get(
                    type_text, _RelatedType.PARALLEL_HISTORY
                )
            elif tag == "a":
                raw_related_id = child_element.get("href")
                related_id = raw_related_id.split("/")[-2]
                related_title = child_element.get_text(strip=True).split(" (")[
                    0
                ]
                all_related_info.append(
                    RelatedInfo(
                        id=related_id, title=related_title, type=last_type
                    )
                )

        page.close()
        if not external_browser:
            browser.__exit__(None, None, None)

        return AnimeInfo(
            id=anime_id,
            title=title,
            poster=poster,
            synopsis=synopsis,
            type=type_,
            rating=None,
            is_finished=is_finished,
            other_titles=all_titles,
            genres=[g.text for g in genres],
            related_info=all_related_info,
            next_episode_date=parsed_date,
            episodes=all_episodes,
        )

    def get_new_episodes(
        self,
        anime_id: str,
        last_episode_number: int,
        tab_timeout: int = 200,
        browser: Optional[SyncBrowser] = None,
    ) -> List[EpisodeInfo]:
        """
        Get the new episodes for an anime.

        Parameters
        ----------
        anime_id : str
            The id of the anime.
        last_episode_number : int
            The last episode number to get.
        tab_timeout : int, optional
            The timeout for waiting for the tab to load. Defaults to 200.
        browser : SyncBrowser, optional
            The browser to use for scraping. If not provided, a new browser
            will be created.

        Returns
        -------
        List[EpisodeInfo]
            A list of new episode information.
        """

        url = f"{BASE_URL}/{anime_id}"
        self._log(f"Getting new episodes for anime with id '{anime_id}'")

        external_browser = browser is not None
        if not external_browser:
            browser = SyncBrowser().__enter__()

        page = browser.new_page()
        page.goto(url)

        select = page.query_selector("div.nice-select.anime__pagination")
        paged_episodes = select.query_selector_all("ul > li")

        all_episodes = []
        finished = False
        for paged_episode in reversed(paged_episodes):
            select.click()

            page.wait_for_timeout(1000)

            current_tabs = browser.context.pages

            if len(current_tabs) > 1:
                current_tabs[-1].close()
                select.click()

            paged_episode.click()

            page.wait_for_timeout(tab_timeout)

            html_text = page.content()
            soup = BeautifulSoup(html_text, "lxml")
            episodes_container = soup.select_one("div#episodes-content")
            episodes = episodes_container.select("div.epcontent")
            for episode in reversed(episodes):
                episode_number = episode.select_one("a")["href"].split("/")[-2]
                image_preview = episode.select_one("a > div")["data-setbg"]
                number = int(episode_number)
                if number <= last_episode_number:
                    finished = True
                    break
                all_episodes.append(
                    EpisodeInfo(
                        number=number,
                        anime_id=anime_id,
                        image_preview=image_preview,
                    )
                )

            if finished:
                break

        page.close()
        if not external_browser:
            browser.__exit__(None, None, None)

        return list(reversed(all_episodes))

    def get_table_download_links(
        self,
        anime_id: str,
        episode_number: int,
        browser: SyncBrowser = None,
    ) -> EpisodeDownloadInfo:
        """
        Get the table download links for an episode.

        Parameters
        ----------
        anime_id : str
            The id of the anime.
        episode_number : int
            The id of the episode.
        browser : SyncBrowser, optional
            The browser to use for scraping. If not provided, a new browser
            will be created.

        Returns
        -------
        EpisodeDownloadInfo
            An object containing the episode id and download links.
        """

        if episode_number < 0:
            raise ValueError("The variable 'episode_number' must be >= 0")

        url = f"{BASE_URL}/{anime_id}/{episode_number}"
        self._log(
            f"Getting table download links for anime '{anime_id}' "
            + f"episode {episode_number}"
        )

        external_browser = browser is not None
        if not external_browser:
            browser = SyncBrowser().__enter__()

        page = browser.new_page()
        page.goto(url)

        html_text = page.content()
        soup = BeautifulSoup(html_text, "lxml")

        download_container = soup.select_one("div.download.mt-2")
        download_rows = download_container.select("tr")[1:]  # Skip header row

        all_download_links = []
        for row in download_rows:
            cells = row.select("td")
            server = cells[0].text
            link = cells[3].select_one("a")["href"]
            all_download_links.append(
                DownloadLinkInfo(server=server, url=link)
            )

        page.close()
        if not external_browser:
            browser.__exit__(None, None, None)

        return EpisodeDownloadInfo(
            episode_number=episode_number,
            download_links=all_download_links,
        )

    def get_iframe_download_links(
        self,
        anime_id: str,
        episode_number: int,
        browser: Optional[SyncBrowser] = None,
    ) -> EpisodeDownloadInfo:
        """
        Note
        ----
        Not supported yet.

        Get the iframe download links for an episode.

        Parameters
        ----------
        anime_id : str
            The id of the anime.
        episode_number : int
            The id of the episode.
        browser : AsyncBrowser, optional
            The browser to use for scraping. If not provided, a new browser
            will be created.

        Returns
        -------
        EpisodeDownloadInfo
            An object containing the episode id and download links.

        Raises
        ------
        TypeError
            If the anime_id or episode_number is not a string or int.
        ValueError
            If the episode_number is less than 0.
        ScraperBlockedError
            If the request is blocked by the server.
        ScraperTimeoutError
            If the request times out.
        ScraperParseError
            If the response from the server cannot be parsed.
        """
        pass

    def get_file_download_link(
        self,
        download_info: DownloadLinkInfo,
        browser: Optional[SyncBrowser] = None,
    ) -> Optional[str]:
        """
        Get the file download link for a download link info object.

        Parameters
        ----------
        download_info : DownloadLinkInfo
            The download link info object.
        browser : SyncBrowser, optional
            The browser to use for scraping. If not provided, a new browser
            will be created.

        Returns
        -------
        str
            The file download link.
        """

        server = download_info.server
        url = download_info.url

        self._log(f"Getting file download link for server '{server}'")

        if server not in get_file_download_link:
            self._log(
                f"Server '{server}' not supported for file download", "ERROR"
            )
            return None

        external_browser = browser is not None
        if not external_browser:
            browser = SyncBrowser().__enter__()

        page = browser.new_page()

        try:
            get_file_fn = get_file_download_link[server]
            file_link = get_file_fn(page, url)

            page.close()
            if not external_browser:
                browser.__exit__(None, None, None)

            return file_link

        except TimeoutError as e:
            self._log("Timeout getting file download link", "ERROR")
            raise ScraperTimeoutError(e)
        except Exception as e:
            self._log("Error getting file download link", "ERROR")
            raise e



from playwright.async_api import Page, TimeoutError as PlaywrightTimeoutError
from ani_scrapy.core.constants.animeflv import (
    SW_DOWNLOAD_URL,
)
from ani_scrapy.core.constants.general import (
    SW_TIMEOUT,
    YOURUPLOAD_TIMEOUT,
)


async def get_sw_file_link(page: Page, url: str):
    await page.goto(url)
    video_id = url.split("/")[-1]
    try_urls = [
        f"{SW_DOWNLOAD_URL}/{video_id}_h",
        f"{SW_DOWNLOAD_URL}/{video_id}_n",
        f"{SW_DOWNLOAD_URL}/{video_id}_l",
    ]

    for url in try_urls:
        retries = 3
        for _ in range(retries):
            try:
                await page.goto(url)

                download_button = await page.wait_for_selector(
                    "form#F1 button", timeout=3000
                )
                await download_button.click()

                try:
                    error_label = await page.wait_for_selector(
                        "div.text-danger.text-center.mb-5", timeout=SW_TIMEOUT
                    )
                    text_label = await error_label.inner_text()
                    if text_label.strip() == "Downloads disabled 620":
                        continue
                    else:
                        break
                except PlaywrightTimeoutError:
                    pass

                download_link = await page.wait_for_selector(
                    "div.text-center a.btn", timeout=SW_TIMEOUT
                )
                download_link = await download_link.get_attribute("href")

                return download_link
            except Exception:
                break
    return None


async def get_yourupload_file_link(page: Page, url: str):
    await page.goto(url)

    try:
        video_element = await page.wait_for_selector(
            "div.jw-media video.jw-video", timeout=YOURUPLOAD_TIMEOUT
        )
    except PlaywrightTimeoutError:
        return None
    video_src = await video_element.get_attribute("src")

    return video_src


async def get_stape_file_link(page: Page, url: str):
    await page.goto(url)

    video_element = await page.wait_for_selector(
        "div.plyr__video-wrapper video", timeout=YOURUPLOAD_TIMEOUT
    )
    video_src = await video_element.get_attribute("src")

    return f"https:{video_src}"



from abc import abstractmethod
from typing import List, Optional

from ani_scrapy.core.base import BaseScraper
from ani_scrapy.core.schemas import (
    AnimeInfo,
    DownloadLinkInfo,
    EpisodeDownloadInfo,
    EpisodeInfo,
    PagedSearchAnimeInfo,
)
from ani_scrapy.sync_api.browser import SyncBrowser


class SyncBaseScraper(BaseScraper):
    """
    Abstract base class for sync anime scrapers.
    """

    @abstractmethod
    def search_anime(self, query: str, **kwargs) -> PagedSearchAnimeInfo:
        pass

    @abstractmethod
    def get_anime_info(
        self,
        anime_id: str,
        include_episodes: bool = True,
        tab_timeout: int = 200,
        **kwargs,
    ) -> AnimeInfo:
        pass

    @abstractmethod
    def get_new_episodes(
        self,
        anime_id: str,
        last_episode_number: int,
        tab_timeout: int = 200,
        browser: Optional[SyncBrowser] = None,
    ) -> List[EpisodeInfo]:
        pass

    @abstractmethod
    def get_table_download_links(
        self, anime_id: str, episode_number: int, **kwargs
    ) -> EpisodeDownloadInfo:
        pass

    @abstractmethod
    def get_iframe_download_links(
        self,
        anime_id: str,
        episode_number: int,
        browser: Optional[SyncBrowser] = None,
    ) -> EpisodeDownloadInfo:
        pass

    @abstractmethod
    def get_file_download_link(
        self,
        download_info: DownloadLinkInfo,
        browser: Optional[SyncBrowser] = None,
    ) -> Optional[str]:
        pass



from ani_scrapy.async_api.scrapers.animeflv.scraper import AnimeFLVScraper
from ani_scrapy.async_api.scrapers.jkanime.scraper import JKAnimeScraper
from ani_scrapy.async_api.browser import AsyncBrowser



from ani_scrapy.core.schemas import _AnimeType, _RelatedType


BASE_URL = "https://animeflv.net"
SEARCH_URL = "https://animeflv.net/browse"
ANIME_VIDEO_URL = "https://animeflv.net/ver"
ANIME_URL = "https://animeflv.net/anime"
BASE_EPISODE_IMG_URL = "https://cdn.animeflv.net/screenshots"

SW_DOWNLOAD_URL = "https://hgplaycdn.com/f"


related_type_map = {
    "Precuela": _RelatedType.PREQUEL,
    "Sequel": _RelatedType.SEQUEL,
    "Historia Paralela": _RelatedType.PARALLEL_HISTORY,
    "Historia Principal": _RelatedType.MAIN_HISTORY,
}

anime_type_map = {
    "Anime": _AnimeType.TV,
    "Pelicula": _AnimeType.MOVIE,
    "OVA": _AnimeType.OVA,
    "Especial": _AnimeType.SPECIAL,
}



from abc import ABC
from loguru import logger
import sys


class BaseScraper(ABC):
    """
    Base class for anime scrapers.
    """

    def __init__(self, verbose: bool = False, level: str = "INFO") -> None:
        self.verbose = verbose
        self.level = level.upper()

        logger.remove()

        if self.verbose:
            logger.add(
                sink=sys.stderr,  # usa stderr para que conserve colores
                level=self.level,
                backtrace=True,
                diagnose=False,
                format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> "
                "| <level>{level: <8}</level> "
                "| <cyan>{name}</cyan>:<cyan>{function}</cyan>:"
                "<cyan>{line}</cyan> "
                "- <level>{message}</level>",
            )

    def _log(self, message: str, level: str = "INFO") -> None:
        """
        Log a message at the given level, showing the correct caller line.
        """
        if not self.verbose:
            return
        log_fn = getattr(logger.opt(depth=1), level.lower(), logger.info)
        log_fn(message)



class ScraperError(Exception):
    """Base exception for all scraping-related errors."""


class ScraperTimeoutError(ScraperError):
    """Failed to fetch due to a timeout."""


class ScraperParseError(ScraperError):
    """Invalid HTML or unexpected structure."""


class ScraperBlockedError(ScraperError):
    """The site blocked the IP (bot detection)."""



from playwright.async_api import Page
from ani_scrapy.core.constants.animeflv import (
    SW_DOWNLOAD_URL,
)
from ani_scrapy.core.constants.general import (
    YOURUPLOAD_DOWNLOAD_URL,
    YOURUPLOAD_TIMEOUT,
    SW_TIMEOUT,
)


async def get_sw_link(page: Page):
    video_element = await page.wait_for_selector(
        "div#video_box", timeout=SW_TIMEOUT
    )
    iframe_element = await video_element.query_selector("iframe")
    iframe_src = await iframe_element.get_attribute("src")
    video_id = iframe_src.split("/")[-1].split("?")[0]
    return f"{SW_DOWNLOAD_URL}/{video_id}"


async def get_yourupload_link(page: Page):
    video_element = await page.wait_for_selector(
        "div#video_box", timeout=YOURUPLOAD_TIMEOUT
    )
    iframe_element = await video_element.query_selector("iframe")
    iframe_src = await iframe_element.get_attribute("src")
    video_id = iframe_src.split("/")[-1].split("?")[0]
    return f"{YOURUPLOAD_DOWNLOAD_URL}/{video_id}"



import re


def clean_text(text: str) -> str:
    text = re.sub(r"[\r\n]+", " ", text)
    text = re.sub(r"\s+", " ", text)
    return text.strip()


def clean_related_type(text: str) -> str:
    text = re.sub(r"[\(\)]", "", text)
    return text.strip()



import json
import requests
from datetime import datetime
from bs4 import BeautifulSoup, Tag
from playwright.sync_api import TimeoutError as PlaywrightTimeoutError

from ani_scrapy.core.constants.animeflv import (
    ANIME_URL,
    ANIME_VIDEO_URL,
    BASE_EPISODE_IMG_URL,
    BASE_URL,
    SEARCH_URL,
)
from ani_scrapy.core.exceptions import (
    ScraperBlockedError,
    ScraperParseError,
    ScraperTimeoutError,
)
from ani_scrapy.core.schemas import (
    _AnimeType,
    _RelatedType,
    AnimeInfo,
    DownloadLinkInfo,
    EpisodeDownloadInfo,
    EpisodeInfo,
    PagedSearchAnimeInfo,
    RelatedInfo,
    SearchAnimeInfo,
)
from ani_scrapy.core.utils.animeflv import get_order_idx
from ani_scrapy.core.utils.general import clean_related_type, clean_text
from ani_scrapy.sync_api.base import SyncBaseScraper
from ani_scrapy.sync_api.browser import SyncBrowser
from typing import List, Optional

from ani_scrapy.sync_api.scrapers.animeflv.file_link import (
    get_sw_file_link,
    get_yourupload_file_link,
)
from ani_scrapy.sync_api.scrapers.animeflv.tab_link import (
    get_sw_link,
    get_yourupload_link,
)
from ani_scrapy.sync_api.scrapers.animeflv.utils import (
    close_not_allowed_popups,
)

related_type_map = {
    "Precuela": _RelatedType.PREQUEL,
    "Sequel": _RelatedType.SEQUEL,
    "Historia Paralela": _RelatedType.PARALLEL_HISTORY,
    "Historia Principal": _RelatedType.MAIN_HISTORY,
}

anime_type_map = {
    "Anime": _AnimeType.TV,
    "Pelicula": _AnimeType.MOVIE,
    "OVA": _AnimeType.OVA,
    "Especial": _AnimeType.SPECIAL,
}

get_tab_download_link = {
    "SW": get_sw_link,
    "YourUpload": get_yourupload_link,
}

get_file_download_link = {
    "SW": get_sw_file_link,
    "YourUpload": get_yourupload_file_link,
}


class AnimeFlvScraper(SyncBaseScraper):
    def _parse_anime_info(self, element: Tag) -> SearchAnimeInfo:
        try:
            anime_id = element.select_one("a[href]")["href"].split("/")[-1]
            type_ = element.select_one("span.Type").text
            title = element.select_one("h3").text
            poster = element.select_one("img")["src"]

            self._log(f"Found anime '{title}' with id '{anime_id}'", "DEBUG")

            return SearchAnimeInfo(
                id=anime_id,
                title=title,
                poster=poster,
                type=anime_type_map.get(type_, _AnimeType.TV),
            )
        except Exception as e:
            raise ScraperParseError(e)

    def search_anime(self, query: str, page: int = 1) -> PagedSearchAnimeInfo:
        """
        Search for anime by query.

        Parameters
        ----------
        query : str
            The query to search for.
        page : int
            The page number to search for. Defaults to 1.

        Returns
        -------
        PagedSearchAnimeInfo
            An object containing the page number, total pages, and animes found.

        Raises
        ------
        ValueError
            If the query is not provided or is less than 3 characters.
        ScraperBlockedError
            If the request is blocked by the server.
        ScraperTimeoutError
            If the request times out.
        ScraperParseError
            If the response from the server cannot be parsed.
        """
        if page < 1:
            raise ValueError("The variable 'page' must be greater than 0")

        if len(query) < 3:
            raise ValueError(
                "The variable 'query' must be at least 3 characters"
            )

        self._log(f"Searching for anime with query '{query}' and page {page}")

        params = {"q": query, "page": page}

        response = requests.get(SEARCH_URL, params=params)

        if response.status_code == 403:
            raise ScraperBlockedError(
                f"Request failed with status code {response.status_code}"
            )
        if response.status_code == 500:
            raise ScraperTimeoutError(
                f"Request failed with status code {response.status_code}"
            )

        soup = BeautifulSoup(response.text, "lxml")
        elements = soup.select("div.Container ul.ListAnimes li article")

        self._log(f"Found {len(elements)} animes")

        animes_info = [self._parse_anime_info(element) for element in elements]

        pagination_links = soup.select("div.NvCnAnm li a")
        total_pages = 1
        if len(pagination_links) > 1:
            total_pages = int(pagination_links[-2].text)

        return PagedSearchAnimeInfo(
            page=page,
            total_pages=total_pages,
            animes=animes_info,
        )

    def get_anime_info(
        self, anime_id: str, include_episodes: bool = True
    ) -> AnimeInfo:
        """
        Get information about an anime.

        Parameters
        ----------
        anime_id : str
            The id of the anime to get information about.

        Returns
        -------
        AnimeInfo
            Information about the anime.

        Raises
        ------
        TypeError
            If the anime_id is not a string.
        ScraperBlockedError
            If the request is blocked by the server.
        ScraperTimeoutError
            If the request times out.
        ScraperParseError
            If the response from the server cannot be parsed.
        """

        if not isinstance(anime_id, str):
            raise TypeError("The variable 'anime_id' must be a string")

        self._log(f"Getting anime info for anime with id '{anime_id}'")

        url = f"{ANIME_URL}/{anime_id}"
        response = requests.get(url)

        if response.status_code == 403:
            raise ScraperBlockedError(
                f"Request failed with status code {response.status_code}"
            )
        if response.status_code == 500:
            raise ScraperTimeoutError(
                f"Request failed with status code {response.status_code}"
            )

        soup = BeautifulSoup(response.text, "lxml")

        title = soup.select_one("h1.Title").text
        poster = soup.select_one("figure img")["src"]
        synopsis = soup.select_one("div.Description p").text
        genres_list = soup.select("nav.Nvgnrs a")
        other_titles_list = soup.select("div.Ficha span.TxtAlt")

        related_info_list = soup.select("ul.ListAnmRel li a")
        related_info = [
            RelatedInfo(
                id=rel["href"].split("/")[-1],
                title=rel.text,
                type=related_type_map.get(
                    clean_related_type(rel.next_sibling),
                    _RelatedType.PARALLEL_HISTORY,
                ),
            )
            for rel in related_info_list
        ]

        type_ = soup.select_one("div.Ficha span.Type").text

        info_ids = []
        episodes_data = []
        episodes = []

        if include_episodes:
            for script in soup.find_all("script"):
                contents = str(script)

                if "var anime_info = [" in contents:
                    anime_info = contents.split("var anime_info = ")[1].split(
                        ";"
                    )[0]
                    info_ids = json.loads(anime_info)

                if "var episodes = [" in contents:
                    data = contents.split("var episodes = ")[1].split(";")[0]
                    episodes_data.extend(json.loads(data))

            anime_thumb_id = info_ids[0]

            for episode_number, _ in reversed(episodes_data):
                number = int(episode_number)
                image_prev = (
                    f"{BASE_EPISODE_IMG_URL}/{anime_thumb_id}/"
                    + f"{number}/th_3.jpg"
                )
                episodes.append(
                    EpisodeInfo(
                        number=number,
                        anime_id=anime_id,
                        image_preview=image_prev,
                    )
                )

        rating = soup.select_one("div.Ficha span.vtprmd").text
        is_finished = (
            soup.select_one("aside.SidebarA span.fa-tv").text == "Finalizado"
        )
        next_episode_date = None
        if len(info_ids) > 3:
            next_episode_date = info_ids[3]

        return AnimeInfo(
            id=anime_id,
            title=title,
            poster=f"{BASE_URL}{poster}",
            synopsis=clean_text(synopsis) if synopsis else None,
            rating=rating,
            is_finished=is_finished,
            type=anime_type_map.get(type_, _AnimeType.TV),
            other_titles=[title.text for title in other_titles_list],
            genres=[genre.text for genre in genres_list],
            related_info=related_info,
            episodes=episodes,
            next_episode_date=(
                datetime.fromisoformat(next_episode_date).date()
                if next_episode_date
                else None
            ),
        )

    def get_new_episodes(
        self,
        anime_id: str,
        last_episode_number: int,
    ) -> List[EpisodeInfo]:
        self._log(f"Getting anime info for anime with id '{anime_id}'")

        url = f"{ANIME_URL}/{anime_id}"
        response = requests.get(url)

        if response.status_code == 403:
            raise ScraperBlockedError(
                f"Request failed with status code {response.status_code}"
            )
        if response.status_code == 500:
            raise ScraperTimeoutError(
                f"Request failed with status code {response.status_code}"
            )

        soup = BeautifulSoup(response.text, "lxml")

        info_ids = []
        episodes_data = []
        episodes = []
        for script in soup.find_all("script"):
            contents = str(script)

            if "var anime_info = [" in contents:
                anime_info = contents.split("var anime_info = ")[1].split(";")[
                    0
                ]
                info_ids = json.loads(anime_info)

            if "var episodes = [" in contents:
                data = contents.split("var episodes = ")[1].split(";")[0]
                episodes_data.extend(json.loads(data))

        anime_thumb_id = info_ids[0]

        for episode_number, _ in episodes_data:
            number = int(episode_number)
            if number <= last_episode_number:
                break
            image_prev = (
                f"{BASE_EPISODE_IMG_URL}/{anime_thumb_id}/{number}"
                + "/th_3.jpg"
            )
            episodes.append(
                EpisodeInfo(
                    number=number,
                    anime_id=anime_id,
                    image_preview=image_prev,
                )
            )
        return list(reversed(episodes))

    def get_table_download_links(
        self,
        anime_id: str,
        episode_number: int,
    ) -> "EpisodeDownloadInfo":
        """
        Get the table download links for an episode.

        Parameters
        ----------
        anime_id : str
            The id of the anime.
        episode_number : int
            The id of the episode.

        Returns
        -------
        EpisodeDownloadInfo
            An object containing the episode id and download links.

        Raises
        ------
        TypeError
            If the anime_id or episode_number is not a string or int.
        ValueError
            If the episode_number is less than 0.
        ScraperBlockedError
            If the request is blocked by the server.
        ScraperTimeoutError
            If the request times out.
        ScraperParseError
            If the response from the server cannot be parsed.
        """
        if episode_number < 0:
            raise ValueError(
                "The variable 'episode_number' must be greater "
                + "than or equal to 0"
            )

        self._log(
            f"Getting table download links for anime with id '{anime_id}' "
            + f"and episode id '{episode_number}'",
        )

        url = f"{ANIME_VIDEO_URL}/{anime_id}-{episode_number}"
        response = requests.get(url)

        if response.status_code == 403:
            raise ScraperBlockedError(
                f"Request failed with status code {response.status_code}"
            )
        if response.status_code == 500:
            raise ScraperTimeoutError(
                f"Request failed with status code {response.status_code}"
            )

        soup = BeautifulSoup(response.text, "lxml")

        rows_list = soup.select("table.RTbl.Dwnl tbody tr")
        rows = []
        for row in rows_list:
            cells = row.select("td")
            self._log(
                f"Found download link for server '{cells[0].text}'",
                "DEBUG",
            )

            rows.append(
                DownloadLinkInfo(
                    server=cells[0].text,
                    url=cells[3].select_one("a")["href"],
                )
            )

        return EpisodeDownloadInfo(
            episode_number=episode_number,
            download_links=rows,
        )

    def get_iframe_download_links(
        self,
        anime_id: str,
        episode_number: int,
        tab_timeout: int = 200,
        browser: Optional[SyncBrowser] = None,
    ) -> "EpisodeDownloadInfo":
        """
        Get the iframe download links for an episode.

        Parameters
        ----------
        anime_id : str
            The id of the anime.
        episode_number : int
            The id of the episode.
        tab_timeout : int, optional
            The timeout for waiting for the tab to load. Defaults to 200.
        browser : Browser, optional
            The browser to use for scraping. If not provided, a new browser
            will be created.

        Returns
        -------
        EpisodeDownloadInfo
            An object containing the episode id and download links.
        """
        url = f"{ANIME_VIDEO_URL}/{anime_id}-{episode_number}"

        self._log(
            f"Getting iframe download links for anime with id '{anime_id}' "
            + f"and episode id '{episode_number}'",
        )

        external_browser = browser is not None
        if not external_browser:
            browser = SyncBrowser().__enter__()

        page = browser.new_page()
        page.on("popup", close_not_allowed_popups)
        page.goto(url)

        server_urls = page.query_selector_all("div.CpCnA ul.CapiTnv li")
        server_names = [
            server_url.get_attribute("title") for server_url in server_urls
        ]

        download_links = []

        order_idx = get_order_idx(server_names)
        for idx in order_idx:
            name = server_names[idx]

            if name not in get_tab_download_link:
                continue

            server_urls[idx].click()
            page.wait_for_timeout(tab_timeout)
            server_urls[idx].click()

            try:
                get_fn = get_tab_download_link[name]
                download_link = get_fn(page)  # ahora sync

                if download_link is None:
                    download_links.append(
                        DownloadLinkInfo(
                            server=name,
                            url=None,
                        )
                    )
                    continue

                download_links.append(
                    DownloadLinkInfo(
                        server=name,
                        url=download_link,
                    )
                )

            except PlaywrightTimeoutError as e:
                self._log("Timeout getting download link", "ERROR")
                download_links.append(
                    DownloadLinkInfo(
                        server=name,
                        url=None,
                    )
                )
                raise ScraperTimeoutError(e)
            except Exception as e:
                self._log("Error getting download link", "ERROR")
                download_links.append(
                    DownloadLinkInfo(
                        server=name,
                        url=None,
                    )
                )
                raise e

        page.close()

        if not external_browser:
            browser.__exit__(None, None, None)

        return EpisodeDownloadInfo(
            episode_number=episode_number,
            download_links=download_links,
        )

    def get_file_download_link(
        self,
        download_info: DownloadLinkInfo,
        browser: Optional[SyncBrowser] = None,
    ) -> Optional[str]:
        """
        Get the file download link for a download link info object.

        Parameters
        ----------
        download_info : DownloadLinkInfo
            The download link info object.
        browser : SyncBrowser, optional
            The browser to use for scraping. If not provided, a new browser
            will be created.

        Returns
        -------
        str | None
            The file download link if found, else None.

        Raises
        ------
        TypeError
            If download_info is not a DownloadLinkInfo object.
        ScraperTimeoutError
            If the operation times out.
        Exception
            For any other unexpected errors.
        """

        if not isinstance(download_info, DownloadLinkInfo):
            raise TypeError("download_info must be a DownloadLinkInfo object")

        server = download_info.server
        url = download_info.url

        self._log(f"Getting file download link for server '{server}'")

        if server not in get_file_download_link:
            self._log(
                f"Server '{server}' not supported for file download", "ERROR"
            )
            return None

        external_browser = browser is not None
        if not external_browser:
            browser = SyncBrowser().__enter__()

        page = None
        try:
            page = browser.new_page()

            page.on("popup", close_not_allowed_popups)

            get_file_fn = get_file_download_link[server]
            file_link = get_file_fn(page, url)

            return file_link

        except PlaywrightTimeoutError as e:
            self._log("Timeout getting file download link", "ERROR")
            raise ScraperTimeoutError(e) from e
        except Exception as e:
            self._log(f"Error getting file download link: {e}", "ERROR")
            raise
        finally:
            if page:
                page.close()
            if not external_browser and browser:
                browser.__exit__(None, None, None)



from ani_scrapy.core.schemas import _AnimeType, _RelatedType


BASE_URL = "https://jkanime.net"
SEARCH_URL = "https://jkanime.net/buscar"
BASE_EPISODE_IMG_URL = (
    "https://cdn.jkdesu.com/assets/images/animes/video/image_thumb"
)

SW_DOWNLOAD_URL = "https://flaswish.com/f"

IMPERSONATE = "chrome"

related_type_map = {
    "Adicional": _RelatedType.PARALLEL_HISTORY,
    "Resumen": _RelatedType.PARALLEL_HISTORY,
    "Version Alternativa": _RelatedType.PARALLEL_HISTORY,
    "Personaje Incluido": _RelatedType.PARALLEL_HISTORY,
    "Secuela": _RelatedType.SEQUEL,
    "Precuela": _RelatedType.PREQUEL,
}


anime_type_map = {
    "Serie": _AnimeType.TV,
    "Pelicula": _AnimeType.MOVIE,
    "OVA": _AnimeType.OVA,
    "Especial": _AnimeType.SPECIAL,
}

month_map = {
    "Enero": 1,
    "Febrero": 2,
    "Marzo": 3,
    "Abril": 4,
    "Mayo": 5,
    "Junio": 6,
    "Julio": 7,
    "Agosto": 8,
    "Septiembre": 9,
    "Octubre": 10,
    "Noviembre": 11,
    "Diciembre": 12,
}

supported_servers = [
    "Streamwish",
    "Mediafire",
]



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional
from datetime import datetime


class _AnimeType(Enum):
    TV = "TV"
    MOVIE = "Movie"
    OVA = "OVA"
    SPECIAL = "Special"


class _RelatedType(Enum):
    PREQUEL = "Prequel"
    SEQUEL = "Sequel"
    PARALLEL_HISTORY = "Parallel History"
    MAIN_HISTORY = "Main History"


@dataclass
class BaseAnimeInfo:
    id: str
    title: str
    type: _AnimeType
    poster: str


@dataclass
class SearchAnimeInfo(BaseAnimeInfo):
    pass


@dataclass
class PagedSearchAnimeInfo:
    page: int
    total_pages: int
    animes: List[SearchAnimeInfo]


@dataclass
class RelatedInfo:
    id: str
    title: str
    type: _RelatedType


@dataclass
class EpisodeInfo:
    number: int
    anime_id: str
    image_preview: Optional[str] = None


@dataclass
class AnimeInfo(BaseAnimeInfo):
    synopsis: str
    is_finished: bool
    rating: Optional[str] = None
    other_titles: List[str] = field(default_factory=list)
    genres: List[str] = field(default_factory=list)
    related_info: List[RelatedInfo] = field(default_factory=list)
    next_episode_date: Optional[datetime] = None
    episodes: List[Optional[EpisodeInfo]] = field(default_factory=list)


@dataclass
class DownloadLinkInfo:
    server: str
    url: Optional[str] = None


@dataclass
class EpisodeDownloadInfo:
    episode_number: int
    download_links: List[DownloadLinkInfo]



from playwright.sync_api import Page

allowed_popups = [
    "www.yourupload.com",
]


def close_not_allowed_popups(page: Page):
    try:
        page.wait_for_load_state("domcontentloaded")
        allowed = False
        for allowed_popup in allowed_popups:
            if allowed_popup in page.url:
                allowed = True
                break

        if not allowed:
            page.close()
    except Exception:
        try:
            page.close()
        except Exception:
            pass



preference_order_tabs = [
    "SW",
    "YourUpload",
]


def get_order_idx(tabs: list[str]) -> list[int]:
    current_tabs = {tab: idx for idx, tab in enumerate(tabs)}

    order_idx = []
    for tab in preference_order_tabs:
        if tab in current_tabs:
            order_idx.append(current_tabs[tab])

    return order_idx



from playwright.async_api import async_playwright
from playwright_stealth import Stealth


from ani_scrapy.core.constants.general import CONTEXT_OPTIONS

stealth = Stealth()


class AsyncBrowser:
    """
    A class for managing an asynchronous browser instance.

    Attributes
    ----------
    headless : bool, optional
        Whether to run the browser in headless mode. Defaults to True.
    executable_path : str, optional
        The path to the browser executable. If not provided, the default
        browser will be chromium.
    args : list[str], optional
        Additional arguments to pass to the browser. Defaults to an empty list.
    """

    def __init__(
        self,
        headless: bool = True,
        executable_path: str | None = None,
        args: list[str] = [],
    ):
        self.headless = headless
        self.executable_path = executable_path
        self.args = args
        self.playwright = None
        self.browser = None

    async def __aenter__(self):
        self.playwright = await async_playwright().start()
        launch_options = {
            "headless": self.headless,
            "args": [
                "--disable-blink-features=AutomationControlled",
                "--no-sandbox",
                *self.args,
            ],
        }
        if self.executable_path:
            launch_options["executable_path"] = self.executable_path
        self.browser = await self.playwright.chromium.launch(**launch_options)
        self.context = await self.browser.new_context(**CONTEXT_OPTIONS)
        await stealth.apply_stealth_async(self.context)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.context.close()
        await self.browser.close()
        await self.playwright.stop()

    async def new_page(self):
        """
        Create a new page in the browser.

        Returns
        -------
        Page
            The new page.
        """
        page = await self.context.new_page()
        return page



CONTEXT_OPTIONS = {
    "ignore_https_errors": True,
    "color_scheme": "dark",
    "locale": "es-EC",
    "no_viewport": True,
    "timezone_id": "America/Guayaquil",
    "user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
    + "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/139.0.0.0 "
    + "Safari/537.36 Edg/139.0.0.0",
    "extra_http_headers": {
        "accept-language": "es-EC",
        "sec-ch-ua": '"Chromium";v="139", "Not;A=Brand";v="99"',
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-platform": '"Windows"',
    },
    "ignore_https_errors": True,
    "viewport": {"width": 1080, "height": 720},
}


STAPE_DOWNLOAD_URL = "https://streamtape.com/get_video"
YOURUPLOAD_DOWNLOAD_URL = "https://yourupload.com/embed"

YOURUPLOAD_TIMEOUT = 10000
SW_TIMEOUT = 7000
MEDIAFIRE_TIMEOUT = 10000
